@@ -13,6 +13,19 @@ from wimpiggy.log import Logger
 log = Logger()
 HOME = os.path.expanduser("~/")
 
+def getoutput(cmd, env=None):
+    try:
+        process = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env, close_fds=True)
+    except Exception, e:
+        print("error running %s: %s" % (cmd, e))
+        raise e
+    (out,err) = process.communicate()
+    code = process.poll()
+    if code!=0:
+        raise Exception("command '%s' returned error code %s, out=%s, err=%s" % (cmd, code, out, err))
+    return out
+
+
 #You will probably need to change those:
 IP = "127.0.0.1"            #this is your IP
 PORT = 10000                #the port to test on
@@ -23,25 +36,28 @@ START_SERVER = True         #if False, you are responsible for starting it
                             #and the data will not be available
 
 SETTLE_TIME = 3             #how long to wait before we start measuring
-MEASURE_TIME = 120           #run for N seconds
+MEASURE_TIME = 120          #run for N seconds
 SERVER_SETTLE_TIME = 3      #how long we wait for the server to start
 DEFAULT_TEST_COMMAND_SETTLE_TIME = 1    #how long we wait after starting the test command
                             #this is the default value, some tests may override this below
 
 TEST_XPRA = True
 TEST_VNC = True
+USE_IPTABLES = True         #this requires iptables to be setup so we can use it for accounting
 
 LIMIT_TESTS = 2
 LIMIT_TESTS = 99999         #to limit the total number of tests being run
+MAX_ERRORS = 100            #allow this many tests to cause errors before aborting
 
 #some commands (games especially) may need longer to startup:
 TEST_COMMAND_SETTLE_TIME = {}
 
-TRICKLE_SHAPING_OPTIONS = [(0, 0, 0)]
-TRICKLE_SHAPING_OPTIONS = [(0, 0, 0), (1024, 1024, 20)]
+NO_SHAPING = (0, 0, 0)
+TRICKLE_SHAPING_OPTIONS = [NO_SHAPING]
+TRICKLE_SHAPING_OPTIONS = [NO_SHAPING, (1024, 1024, 20)]
 TRICKLE_SHAPING_OPTIONS = [(1024, 1024, 20), (128, 32, 40), (0, 0, 0)]
-TRICKLE_SHAPING_OPTIONS = [(0, 0, 0), (1024, 256, 20), (128, 32, 200), (32, 8, 100)]
-
+TRICKLE_SHAPING_OPTIONS = [NO_SHAPING, (1024, 256, 20), (1024, 256, 300), (128, 32, 100), (32, 8, 200)]
+TRICKLE_SHAPING_OPTIONS = [NO_SHAPING, (1024, 256, 20), (256, 64, 50), (128, 32, 100), (32, 8, 200)]
 
 #tools we use:
 IPTABLES_CMD = ["sudo", "/usr/sbin/iptables"]
@@ -58,8 +74,9 @@ GLX_TESTS = [GLX_SPHERES, GLX_GEARS]
 #the plain X11 tests:
 X11_PERF = ["/usr/bin/x11perf", "-resize", "-all"]
 XTERM_TEST = ["/usr/bin/xterm", "-geometry", "160x60", "-e", "while true; do dmesg; done"]
+FAKE_CONSOLE_USER_TEST = ["/usr/bin/xterm", "-geometry", "160x60", "-e", "PYTHONPATH=`pwd` ./tests/xpra/simulate_console_user.py"]
 GTKPERF_TEST = "while true; do gtkperf -a; done"
-X11_TESTS = [X11_PERF, XTERM_TEST, GTKPERF_TEST]
+X11_TESTS = [X11_PERF, XTERM_TEST, FAKE_CONSOLE_USER_TEST, GTKPERF_TEST]
 
 #the screensaver tests:
 XSCREENSAVERS_PATH = "/usr/libexec/xscreensaver"
@@ -123,23 +140,31 @@ VNC_JPEG_OPTIONS = [-1, 4]
 
 
 XPRA_BIN = "/usr/bin/xpra"
-XPRA_SERVER_START_COMMAND = [XPRA_BIN, "--no-daemon", "--bind-tcp=0.0.0.0:%s" % PORT,
-                       "start", ":%s" % DISPLAY_NO,
-                       "--xvfb=%s -nolisten tcp +extension GLX +extension RANDR +extension RENDER -logfile %s -config %s" % (XORG_BIN, XORG_LOG, XORG_CONFIG)]
 XPRA_SERVER_STOP_COMMANDS = [
                              [XPRA_BIN, "stop", ":%s" % DISPLAY_NO],
                              "ps -ef | grep -i [X]org-for-Xpra-:%s | awk '{print $2}' | xargs kill" % DISPLAY_NO
                              ]
 XPRA_INFO_COMMAND = [XPRA_BIN, "info", "tcp:%s:%s" % (IP, PORT)]
-XPRA_TEST_ENCODINGS = ["png", "x264", "mmap"]
-XPRA_TEST_ENCODINGS = ["png", "rgb24", "jpeg", "x264", "vpx", "mmap"]
-XPRA_TEST_ENCODINGS = ["png", "jpeg", "x264", "vpx", "mmap"]
-XPRA_JPEG_OPTIONS = [40, 80, 90]
-XPRA_JPEG_OPTIONS = [40, 90]
-XPRA_JPEG_OPTIONS = [80]
+XPRA_USE_XDUMMY = True
+XPRA_QUALITY_OPTIONS = [40, 90]
+XPRA_QUALITY_OPTIONS = [80]
+XPRA_QUALITY_OPTIONS = [10, 40, 80, 90]
 XPRA_COMPRESSION_OPTIONS = [0, 3, 9]
 XPRA_COMPRESSION_OPTIONS = [0, 3]
-XPRA_COMPRESSION_OPTIONS = [3]
+XPRA_COMPRESSION_OPTIONS = [None]
+XPRA_CONNECT_OPTIONS = ["ssh", "tcp", "unix"]
+XPRA_VERSION = getoutput([XPRA_BIN, "--version"]).replace("\n", "").replace("\r", "").split()[1].replace("v0.", "0.")
+XPRA_VERSION_NO = [int(x) for x in XPRA_VERSION.split(".")]
+print ("XPRA_VERSION_NO=%s" % XPRA_VERSION_NO)
+XPRA_TEST_ENCODINGS = ["png", "x264", "mmap"]
+XPRA_TEST_ENCODINGS = ["png", "jpeg", "x264", "vpx", "mmap"]
+XPRA_TEST_ENCODINGS = ["png", "rgb24", "jpeg", "x264", "vpx", "mmap"]
+if XPRA_VERSION_NO>=[0, 7]:
+    XPRA_TEST_ENCODINGS.append("webp")
+XPRA_ENCODING_QUALITY_OPTIONS = {"jpeg" : XPRA_QUALITY_OPTIONS,
+                                 "webp" : XPRA_QUALITY_OPTIONS,
+                                 "x264" : XPRA_QUALITY_OPTIONS+[-1],
+                                 }
 
 
 check = [TRICKLE_BIN]
@@ -174,18 +199,6 @@ def try_to_kill(process, grace=0):
         except Exception, e:
             print("could not stop process %s: %s" % (process, e))
 
-def getoutput(cmd, env=None):
-    try:
-        process = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env, close_fds=True)
-    except Exception, e:
-        print("error running %s: %s" % (cmd, e))
-        raise e
-    (out,err) = process.communicate()
-    code = process.poll()
-    if code!=0:
-        raise Exception("command '%s' returned error code %s, out=%s, err=%s" % (cmd, code, out, err))
-    return out
-
 def find_matching_lines(out, pattern):
     lines = []
     for line in out.splitlines():
@@ -259,12 +272,9 @@ if err:
         VNCVIEWER_VERSION = "TigerVNC Viewer %s" % (v_lines[0].split()[5])
 print ("VNCVIEWER_VERSION=%s" % VNCVIEWER_VERSION)
 
-XPRA_VERSION = getoutput([XPRA_BIN, "--version"]).replace("\n", "").replace("\r", "").split()[1].replace("v0.", "0.")
-XPRA_VERSION_NO = [int(x) for x in XPRA_VERSION.split(".")]
-print ("XPRA_VERSION_NO=%s" % XPRA_VERSION_NO)
-
 SVN_VERSION = getoutput(["svnversion", "-n"])
 
+WINDOW_MANAGER = os.environ.get("DESKTOP_SESSION", "unknown")
 
 def clean_sys_state():
     #clear the caches
@@ -273,6 +283,8 @@ def clean_sys_state():
     assert process.wait()==0, "failed to run %s" % str(cmd)
 
 def zero_iptables():
+    if not USE_IPTABLES:
+        return
     cmds = [IPTABLES_CMD+['-Z', 'INPUT'], IPTABLES_CMD+['-Z', 'OUTPUT']]
     for cmd in cmds:
         getoutput(cmd)
@@ -322,6 +334,8 @@ def getiptables_line(chain, pattern, setup_info):
     return line
 
 def parse_ipt(chain, pattern, setup_info):
+    if not USE_IPTABLES:
+        return  0, 0
     line = getiptables_line(chain, pattern, setup_info)
     parts = line.split()
     assert len(parts)>2
@@ -396,7 +410,7 @@ def with_server(start_server_command, stop_server_commands, in_tests, get_stats_
     errors = 0
     results = []
     count = 0
-    for name, tech_name, server_version, client_version, encoding, compression, (down,up,latency), test_command, client_cmd in tests:
+    for name, tech_name, server_version, client_version, encoding, compression, ssh, (down,up,latency), test_command, client_cmd in tests:
         try:
             print("**************************************************************")
             count += 1
@@ -429,9 +443,9 @@ def with_server(start_server_command, stop_server_commands, in_tests, get_stats_
                     #run the client test
                     result = [name, tech_name, server_version, client_version, " ".join(sys.argv[1:]), SVN_VERSION]
                     result += [encoding, get_command_name(test_command)]
-                    result += [MEASURE_TIME, time.time(), CPU_INFO, PLATFORM, KERNEL_VERSION, XORG_VERSION, OPENGL_INFO]
+                    result += [MEASURE_TIME, time.time(), CPU_INFO, PLATFORM, KERNEL_VERSION, XORG_VERSION, OPENGL_INFO, WINDOW_MANAGER]
                     result += ["%sx%s" % gdk.get_default_root_window().get_size()]
-                    result += [compression, down, up, latency]
+                    result += [compression, ssh, down, up, latency]
                     result += measure_client(server_pid, name, client_cmd, get_stats_cb)
                     results.append(result)
                 except Exception, e:
@@ -439,7 +453,7 @@ def with_server(start_server_command, stop_server_commands, in_tests, get_stats_
                     traceback.print_exc()
                     errors += 1
                     print("error during client command run for %s: %s" % (name, e))
-                    if errors>10:
+                    if errors>MAX_ERRORS:
                         print("too many errors, aborting tests")
                         break
             finally:
@@ -503,17 +517,13 @@ def get_stats_headers():
              "Min Batch Delay (ms)", "Max Batch Delay (ms)", "Avg Batch Delay (ms)",
              "Application packets in/s", "Application bytes in/s", "Application packets out/s", "Application bytes out/s", "mmap bytes/s",
              "Min Client Latency (ms)", "Max Client Latency (ms)", "Avg Client Latency (ms)",
-             "Min Server Latency (ms)", "Max Server Latency (ms)", "Avg Server Latency (ms)",
+             "Min Client Ping Latency (ms)", "Max Client Ping Latency (ms)", "Avg Client Ping Latency (ms)",
+             "Min Server Ping Latency (ms)", "Max Server Ping Latency (ms)", "Avg Server Ping Latency (ms)",
              "Min Damage Latency (ms)", "Max Damage Latency (ms)", "Avg Damage Latency (ms)",
              ]
 
 def no_stats(last_record=None):
-    return  ["", "", "", "",
-             "", "", "", "", "", "",
-             "", "", "",
-             "", "", "",
-             "", "", "",
-             ]
+    return  ["" for _ in xrange(24)]
 
 def xpra_get_stats(last_record=None):
     if XPRA_VERSION_NO<[0, 3]:
@@ -538,71 +548,99 @@ def xpra_get_stats(last_record=None):
         last_output_packetcount = last_record[index+2]
         last_output_bytecount = last_record[index+3]
         last_mmap_bytes = last_record[index+4]
-    def get(old_var_name, new_var_name=None, default_value=""):
+    def get(names, default_value=""):
         """ some of the fields got renamed, try both old and new names """
-        v = d.get(old_var_name)
-        if not v and new_var_name is not None:
-            v = d.get(new_var_name)
-        if not v:
-            v = default_value
-        return v
+        for n in names:
+            v = d.get(n)
+            if v is not None:
+                return v
+        return default_value
     return [
-            get("regions_per_second"),
-            get("pixels_per_second"),
-            get("pixels_encoded_per_second"),
-            get("pixels_decoded_per_second"),
-            get("min_batch_delay", "batch_delay.min"),
-            get("max_batch_delay", "batch_delay.max"),
-            get("avg_batch_delay", "batch_delay.avg"),
-            (int(get("input_packetcount", None, 0))-last_input_packetcount)/MEASURE_TIME,
-            (int(get("input_bytecount", None, 0))-last_input_bytecount)/MEASURE_TIME,
-            (int(get("output_packetcount", None, 0))-last_output_packetcount)/MEASURE_TIME,
-            (int(get("output_bytecount", None, 0))-last_output_bytecount)/MEASURE_TIME,
-            (int(get("output_mmap_bytecount", None, 0))-last_mmap_bytes)/MEASURE_TIME,
-            get("min_client_latency", "client_latency.min"),
-            get("max_client_latency", "client_latency.max"),
-            get("avg_client_latency", "client_latency.avg"),
-            get("min_server_latency", "server_latency.min"),
-            get("max_server_latency", "server_latency.max"),
-            get("avg_server_latency", "server_latency.avg"),
-            get("damage_in_latency.min"),
-            get("damage_in_latency.max"),
-            get("damage_in_latency.avg"),
+            get(["regions_per_second"]),
+            get(["pixels_per_second"]),
+            get(["pixels_encoded_per_second"]),
+            get(["pixels_decoded_per_second"]),
+            get(["batch_delay.min", "min_batch_delay"]),
+            get(["batch_delay.max", "max_batch_delay"]),
+            get(["batch_delay.avg", "avg_batch_delay"]),
+            (int(get(["input_packetcount"], 0))-last_input_packetcount)/MEASURE_TIME,
+            (int(get(["input_bytecount"], 0))-last_input_bytecount)/MEASURE_TIME,
+            (int(get(["output_packetcount"], 0))-last_output_packetcount)/MEASURE_TIME,
+            (int(get(["output_bytecount"], 0))-last_output_bytecount)/MEASURE_TIME,
+            (int(get(["output_mmap_bytecount"], 0))-last_mmap_bytes)/MEASURE_TIME,
+            get(["client_latency.min", "min_client_latency"]),
+            get(["client_latency.max", "max_client_latency"]),
+            get(["client_latency.avg", "avg_client_latency"]),
+            get(["client_ping_latency.min"]),
+            get(["client_ping_latency.max"]),
+            get(["client_ping_latency.avg"]),
+            get(["server_ping_latency.min", "server_latency.min", "min_server_latency"]),
+            get(["server_ping_latency.max", "server_latency.max", "max_server_latency"]),
+            get(["server_ping_latency.avg", "server_latency.avg", "avg_server_latency"]),
+            get(["damage_in_latency.min"]),
+            get(["damage_in_latency.max"]),
+            get(["damage_in_latency.avg"]),
            ]
 
+def get_xpra_start_server_command():
+    cmd = [XPRA_BIN, "--no-daemon", "--bind-tcp=0.0.0.0:%s" % PORT]
+    if XPRA_USE_XDUMMY:
+        cmd.append("--xvfb=%s -nolisten tcp +extension GLX +extension RANDR +extension RENDER -logfile %s -config %s" % (XORG_BIN, XORG_LOG, XORG_CONFIG))
+    if XPRA_VERSION_NO>=[0, 5]:
+        cmd.append("--no-notifications")
+    cmd += ["start", ":%s" % DISPLAY_NO]
+    return cmd
+
 def test_xpra():
     print("")
     print("*********************************************************")
     print("                Xpra tests")
     print("")
     tests = []
-    for down,up,latency in TRICKLE_SHAPING_OPTIONS:
-        for x11_test_command in X11_TEST_COMMANDS:
-            for encoding in XPRA_TEST_ENCODINGS:
-                QUALITY = [-1]
-                if encoding=="jpeg":
-                    QUALITY = XPRA_JPEG_OPTIONS
-                for jpeg_q in QUALITY:
-                    comp_options = XPRA_COMPRESSION_OPTIONS
-                    for compression in comp_options:
-                        cmd = trickle_command(down, up, latency)
-                        cmd += [XPRA_BIN,
-                               "attach", "tcp:%s:%s" % (IP, PORT),
-                               "-z", str(compression), "--readonly"]
-                        if XPRA_VERSION_NO>=[0, 3]:
-                            cmd.append("--enable-pings")
-                        if encoding=="jpeg":
-                            cmd.append("--jpeg-quality=%s" % jpeg_q)
-                            name = "%s-%s" % (encoding, jpeg_q)
-                        else:
-                            name = encoding
-                        if encoding!="mmap":
-                            cmd.append("--no-mmap")
-                            cmd.append("--encoding=%s" % encoding)
-                        command_name = get_command_name(x11_test_command)
-                        test_name = "%s (%s - %s - %s)" % (name, command_name, compression, trickle_str(down, up, latency))
-                        tests.append((test_name, "xpra", XPRA_VERSION, XPRA_VERSION, encoding, compression, (down,up,latency), x11_test_command, cmd))
-    return with_server(XPRA_SERVER_START_COMMAND, XPRA_SERVER_STOP_COMMANDS, tests, xpra_get_stats)
+    for connect_option in XPRA_CONNECT_OPTIONS:
+        shaping_options = TRICKLE_SHAPING_OPTIONS
+        if connect_option=="unix":
+            shaping_options = [NO_SHAPING]
+        for down,up,latency in shaping_options:
+            for x11_test_command in X11_TEST_COMMANDS:
+                for encoding in XPRA_TEST_ENCODINGS:
+                    quality_options = XPRA_ENCODING_QUALITY_OPTIONS.get(encoding, [-1])
+                    for quality in quality_options:
+                        comp_options = XPRA_COMPRESSION_OPTIONS
+                        for compression in comp_options:
+                            cmd = trickle_command(down, up, latency)
+                            cmd += [XPRA_BIN, "attach"]
+                            if connect_option=="ssh":
+                                cmd.append("ssh:%s:%s" % (IP, DISPLAY_NO))
+                            elif connect_option=="tcp":
+                                cmd.append("tcp:%s:%s" % (IP, PORT))
+                            else:
+                                cmd.append(":%s" % (DISPLAY_NO))
+                            cmd.append("--readonly")
+                            if compression is not None:
+                                cmd += ["-z", str(compression)]
+                            if XPRA_VERSION_NO>=[0, 3]:
+                                cmd.append("--enable-pings")
+                                cmd.append("--no-clipboard")
+                            if XPRA_VERSION_NO>=[0, 5]:
+                                cmd.append("--no-bell")
+                                cmd.append("--no-cursors")
+                                cmd.append("--no-notifications")
+                            if encoding in ("jpeg", "webp"):
+                                if XPRA_VERSION_NO>=[0, 7]:
+                                    cmd.append("--quality=%s" % quality)
+                                else:
+                                    cmd.append("--jpeg-quality=%s" % quality)
+                                name = "%s-%s" % (encoding, quality)
+                            else:
+                                name = encoding
+                            if encoding!="mmap":
+                                cmd.append("--no-mmap")
+                                cmd.append("--encoding=%s" % encoding)
+                            command_name = get_command_name(x11_test_command)
+                            test_name = "%s (%s - %s - %s - via %s)" % (name, command_name, compression, trickle_str(down, up, latency), connect_option)
+                            tests.append((test_name, "xpra", XPRA_VERSION, XPRA_VERSION, encoding, compression, connect_option, (down,up,latency), x11_test_command, cmd))
+    return with_server(get_xpra_start_server_command(), XPRA_SERVER_STOP_COMMANDS, tests, xpra_get_stats)
 
 
 def get_x11_client_window_info(display, *app_name_strings):
@@ -695,11 +733,7 @@ def get_vnc_stats(last_record=None):
                 parts = line.split()
                 regions_s = parts[-1]
                 print("Frames/sec=%s" % regions_s)
-    return  [regions_s, "", "", "",
-             "", "", "", "", "", "",
-             "", "", "",
-             "", "", ""
-             ]
+    return  [regions_s] + ["" for _ in xrange(20)]
 
 def test_vnc():
     print("")
@@ -741,7 +775,7 @@ def test_vnc():
                                 zlibtxt = "zlib=%s" % zlib
                             command_name = get_command_name(x11_test_command)
                             test_name = "vnc (%s - %s - %s - compression=%s - %s - %s)" % (command_name, encoding, zlibtxt, compression, jpegtxt, trickle_str(down, up, latency))
-                            tests.append((test_name, "vnc", XVNC_VERSION, VNCVIEWER_VERSION, encoding, compression, (down,up,latency), x11_test_command, cmd))
+                            tests.append((test_name, "vnc", XVNC_VERSION, VNCVIEWER_VERSION, encoding, compression, False, (down,up,latency), x11_test_command, cmd))
     return with_server(XVNC_SERVER_START_COMMAND, XVNC_SERVER_STOP_COMMANDS, tests, get_vnc_stats)
 
 
@@ -761,8 +795,8 @@ def main():
     print("")
     headers = ["Test Name", "Remoting Tech", "Server Version", "Client Version", "Custom Params", "SVN Version",
                "Encoding", "Test Command", "Sample Duration (s)", "Sample Time (epoch)",
-               "CPU info", "Platform", "Kernel Version", "Xorg version", "OpenGL", "Screen Size",
-               "compression", "download limit (KB)", "upload limit (KB)", "latency (ms)",
+               "CPU info", "Platform", "Kernel Version", "Xorg version", "OpenGL", "Client Window Manager", "Screen Size",
+               "compression", "Connect via", "download limit (KB)", "upload limit (KB)", "latency (ms)",
                "packets in/s", "packets in: bytes/s", "packets out/s", "packets out: bytes/s"]
     headers += get_stats_headers()
     headers += ["client user cpu_pct", "client system cpu pct", "client number of threads", "client vsize (MB)", "client rss (MB)",
