@@ -35,14 +35,15 @@ from wimpiggy.lowlevel import (
                XRemoveFromSaveSet,                          #@UnresolvedImport
                XSetInputFocus,                              #@UnresolvedImport
                XKillClient,                                 #@UnresolvedImport
-               send_wm_take_focus,                          #@UnresolvedImport
-               send_wm_delete_window,                       #@UnresolvedImport
                sendConfigureNotify,                         #@UnresolvedImport
                configureAndNotify,                          #@UnresolvedImport
                substructureRedirect                         #@UnresolvedImport
                )
+from wimpiggy.lowlevel.send_wm import (
+                send_wm_take_focus,                         #@UnresolvedImport
+                send_wm_delete_window)                      #@UnresolvedImport
 from wimpiggy.util import (AutoPropGObjectMixin,
-                           one_arg_signal,
+                           one_arg_signal, no_arg_signal,
                            non_none_list_accumulator)
 from wimpiggy.error import trap, XError
 from wimpiggy.prop import prop_get, prop_set
@@ -247,6 +248,7 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
 
         log("new window %s", hex(client_window.xid))
 
+        self._managed = False
         self._composite = None
         self.client_window = client_window
         self._internal_set_property("client-window", client_window)
@@ -256,16 +258,20 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
             # Keith Packard says that composite state is undefined following a
             # reparent, so I'm not sure doing this here in the superclass,
             # before we reparent, actually works... let's wait and see.
-            self._composite = CompositeHelper(self.client_window, False, )  #self.composite_configure_event)
+            self._composite = CompositeHelper(self.client_window, False)
             h = self._composite.connect("contents-changed", self._forward_contents_changed)
             self._composite.connect("wimpiggy-configure-event", self.composite_configure_event)
             self._damage_forward_handle = h
             self._geometry = geometry_with_border(self.client_window)
+            self._managed = True
         try:
             trap.call(setup)
         except XError, e:
             raise Unmanageable(e)
 
+    def is_managed(self):
+        return self._managed
+
     def _forward_contents_changed(self, obj, event):
         self.emit("client-contents-changed", event)
 
@@ -285,19 +291,24 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
 
     def composite_configure_event(self, composite_window, event):
         log("BaseWindowModel.composite_configure_event(%s,%s)", composite_window, event)
+        if self._composite:
+            self._composite.do_wimpiggy_configure_event(event)
 
     def do_get_property_geometry(self, pspec):
         (x, y, w, h, b) = self._geometry
         return (x, y, w + 2*b, h + 2*b)
 
     def unmanage(self, exiting=False):
-        self.emit("unmanaged", exiting)
+        if self._managed:
+            self._managed = False
+            self.emit("unmanaged", exiting)
 
     def do_unmanaged(self, wm_exiting):
         remove_event_receiver(self.client_window, self)
         if self._composite:
             self._composite.disconnect(self._damage_forward_handle)
             self._composite.destroy()
+            self._composite = None
 
     def _read_initial_properties(self):
         transient_for = self.prop_get("WM_TRANSIENT_FOR", "window")
@@ -343,6 +354,7 @@ class OverrideRedirectWindowModel(BaseWindowModel):
             if not is_mapped(self.client_window):
                 raise Unmanageable("window already unmapped")
 
+            self._managed = True
             self._read_initial_properties()
         try:
             trap.call(setup)
@@ -355,8 +367,15 @@ class OverrideRedirectWindowModel(BaseWindowModel):
     def do_wimpiggy_unmap_event(self, event):
         self.unmanage()
 
-    def prop_get(self, key, type):
-        return prop_get(self.client_window, key, type, ignore_errors=False)
+    def prop_get(self, key, ptype):
+        return prop_get(self.client_window, key, ptype, ignore_errors=False)
+
+    def get_dimensions(self):
+        _, _, ww, wh = self.get_property("geometry")
+        return ww, wh
+
+    def is_OR(self):
+        return  True
 
 
 gobject.type_register(OverrideRedirectWindowModel)
@@ -451,8 +470,8 @@ class WindowModel(BaseWindowModel):
     __gsignals__ = {
         # X11 bell event:
         "bell": one_arg_signal,
-        "geometry": one_arg_signal,
-        
+        "geometry": no_arg_signal,
+
         "ownership-election": (gobject.SIGNAL_RUN_LAST,
                                gobject.TYPE_PYOBJECT, (),
                                non_none_list_accumulator),
@@ -534,12 +553,18 @@ class WindowModel(BaseWindowModel):
             raise Unmanageable(e)
         self._setup_done = True
 
-    def prop_get(self, key, type, ignore_errors=False):
+    def is_OR(self):
+        return  False
+
+    def get_dimensions(self):
+        return  self.get_property("actual-size")
+
+    def prop_get(self, key, ptype, ignore_errors=False):
         # Utility wrapper for prop_get on the client_window
         # also allows us to ignore property errors during setup_client
         if not self._setup_done:
             ignore_errors = True
-        return prop_get(self.client_window, key, type, ignore_errors=ignore_errors)
+        return prop_get(self.client_window, key, ptype, ignore_errors=ignore_errors)
 
     def do_wimpiggy_xkb_event(self, event):
         log("WindowModel.do_wimpiggy_xkb_event(%r)" % event)
@@ -560,7 +585,7 @@ class WindowModel(BaseWindowModel):
         pass
 
     def do_wimpiggy_unmap_event(self, event):
-        if event.delivered_to is self.corral_window:
+        if event.delivered_to is self.corral_window or self.corral_window is None:
             return
         assert event.window is self.client_window
         # The client window got unmapped.  The question is, though, was that
@@ -580,7 +605,7 @@ class WindowModel(BaseWindowModel):
             self.unmanage()
 
     def do_wimpiggy_destroy_event(self, event):
-        if event.delivered_to is self.corral_window:
+        if event.delivered_to is self.corral_window or self.corral_window is None:
             return
         assert event.window is self.client_window
         # This is somewhat redundant with the unmap signal, because if you
@@ -592,7 +617,7 @@ class WindowModel(BaseWindowModel):
         self.unmanage()
 
     def do_unmanaged(self, exiting):
-        log("unmanaging window")
+        log("unmanaging window: %s (%s - %s)", self, self.corral_window, self.client_window)
         self._internal_set_property("owner", None)
         def unmanageit():
             self._scrub_withdrawn_window()
@@ -608,8 +633,10 @@ class WindowModel(BaseWindowModel):
             sendConfigureNotify(self.client_window)
             if exiting:
                 self.client_window.show_unraised()
-        trap.swallow(unmanageit)
-        self.corral_window.destroy()
+        if self.corral_window:
+            trap.swallow(unmanageit)
+            self.corral_window.destroy()
+            self.corral_window = None
         BaseWindowModel.do_unmanaged(self, exiting)
 
     def ownership_election(self):
@@ -640,14 +667,27 @@ class WindowModel(BaseWindowModel):
     def _sanitize_size_hints(self, size_hints):
         if size_hints is None:
             return
-        for attr in ["max_size", "min_size", "base_size",
-                    "resize_inc", "min_aspect", "max_aspect"]:
+        for attr in ["min_aspect", "max_aspect"]:
             #"min_aspect_ratio", "max_aspect_ratio" ??
             v = getattr(size_hints, attr)
             if v is not None:
-                w,h = v
-                if w>=(2**32-1) or h>(2**32-1):
-                    log.info("clearing invalid size hint value for %s: %s", attr, v)
+                try:
+                    f = float(v)
+                except:
+                    f = None
+                if f is None or f>=(2**32-1):
+                    log.warn("clearing invalid aspect hint value for %s: %s", attr, v)
+                    setattr(size_hints, attr, -1.0)
+        for attr in ["max_size", "min_size", "base_size", "resize_inc"]:
+            #"min_aspect_ratio", "max_aspect_ratio" ??
+            v = getattr(size_hints, attr)
+            if v is not None:
+                try:
+                    w,h = v
+                except:
+                    w,h = None,None
+                if (w is None or h is None) or w>=(2**32-1) or h>(2**32-1):
+                    log.warn("clearing invalid size hint value for %s: %s", attr, v)
                     setattr(size_hints, attr, (-1,-1))
 
     def _update_client_geometry(self):
@@ -676,23 +716,45 @@ class WindowModel(BaseWindowModel):
         self._internal_set_property("user-friendly-size", (wvis, hvis))
 
     def composite_configure_event(self, composite_window, event):
+        log("WindowModel.composite_configure_event(%s, %s)", composite_window, event)
+        BaseWindowModel.composite_configure_event(self, composite_window, event)
+        gobject.idle_add(self.may_resize_corral_window)
+
+    def may_resize_corral_window(self):
+        if not self._managed:
+            return
+        if self.corral_window is None or not self.corral_window.is_visible():
+            return
+        if self.client_window is None or not self.client_window.is_visible():
+            return
+        try:
+            #workaround applications whose windows disappear from underneath us:
+            if trap.call(self.resize_corral_window):
+                self.emit("geometry")
+        except XError, e:
+            log.warn("failed to resize corral window: %s", e)
+
+    def resize_corral_window(self):
         #the client window may have been resized (generally programmatically)
         #so we may need to update the corral_window to match
-        log("WindowModel.composite_configure_event(%s, %s)", composite_window, event)
         cow, coh = self.corral_window.get_geometry()[2:4]
-        clw, clh = self.client_window.get_geometry()[2:4]
+        clx, cly, clw, clh = self.client_window.get_geometry()[:4]
+        if (clx, cly) != (0, 0):
+            log("resize_corral_window() client window has moved, resetting it")
+            self.client_window.move(0, 0)
         if cow!=clw or coh!=clh:
-            log("WindowModel.composite_configure_event(%s, %s) corral window (%sx%s) does not match client window (%sx%s), resizing it",
-                     composite_window, event, cow, coh, clw, clh)
+            log("resize_corral_window() corral window (%sx%s) does not match client window (%sx%s), resizing it",
+                     cow, coh, clw, clh)
             self.corral_window.resize(clw, clh)
             hints = self.get_property("size-hints")
             self._sanitize_size_hints(hints)
             size = calc_constrained_size(clw, clh, hints)
-            log("composite_configure_event: new constrained size=%s", size)
+            log("resize_corral_window() new constrained size=%s", size)
             w, h, wvis, hvis = size
             self._internal_set_property("actual-size", (w, h))
             self._internal_set_property("user-friendly-size", (wvis, hvis))
-            self.emit("geometry", event)
+            return True
+        return False
 
     def do_child_configure_request_event(self, event):
         # Ignore the request, but as per ICCCM 4.1.5, send back a synthetic
@@ -768,24 +830,24 @@ class WindowModel(BaseWindowModel):
     _property_handlers["WM_NORMAL_HINTS"] = _handle_wm_normal_hints
 
     def _handle_title_change(self):
-        wm_name = self.prop_get("WM_NAME", "latin1", True)
         net_wm_name = self.prop_get("_NET_WM_NAME", "utf8", True)
         if net_wm_name is not None:
             self._internal_set_property("title", net_wm_name)
         else:
             # may be None
+            wm_name = self.prop_get("WM_NAME", "latin1", True)
             self._internal_set_property("title", wm_name)
 
     _property_handlers["WM_NAME"] = _handle_title_change
     _property_handlers["_NET_WM_NAME"] = _handle_title_change
 
     def _handle_icon_title_change(self):
-        wm_icon_name = self.prop_get("WM_ICON_NAME", "latin1", True)
         net_wm_icon_name = self.prop_get("_NET_WM_ICON_NAME", "utf8", True)
         if net_wm_icon_name is not None:
             self._internal_set_property("icon-title", net_wm_icon_name)
         else:
             # may be None
+            wm_icon_name = self.prop_get("WM_ICON_NAME", "latin1", True)
             self._internal_set_property("icon-title", wm_icon_name)
 
     _property_handlers["WM_ICON_NAME"] = _handle_icon_title_change
@@ -830,7 +892,7 @@ class WindowModel(BaseWindowModel):
     def _read_initial_properties(self):
         # Things that don't change:
         BaseWindowModel._read_initial_properties(self)
-        
+
         geometry = self.client_window.get_geometry()
         self._internal_set_property("requested-position", (geometry[0], geometry[1]))
         self._internal_set_property("requested-size", (geometry[2], geometry[3]))
@@ -1014,6 +1076,8 @@ class WindowModel(BaseWindowModel):
     def give_client_focus(self):
         """The focus manager has decided that our client should recieve X
         focus.  See world_window.py for details."""
+        if self.corral_window is None:
+            return  True
         log("Giving focus to client")
         # Have to fetch the time, not just use CurrentTime, both because ICCCM
         # says that WM_TAKE_FOCUS must use a real time and because there are
