@@ -37,10 +37,16 @@ else:
         for gtkwindow in gtkwindows:
             gtkwindow.get_window().set_cursor(cursor)
 
-
+import sys
 import os
 import time
 import ctypes
+from threading import Thread
+try:
+    from queue import Queue     #@UnresolvedImport @UnusedImport (python3)
+except:
+    from Queue import Queue     #@Reimport
+
 
 from wimpiggy.util import (n_arg_signal,
                            gtk_main_quit_really,
@@ -50,10 +56,12 @@ from wimpiggy.log import Logger
 log = Logger()
 
 from xpra.deque import maxdeque
-from xpra.client_base import XpraClientBase
-from xpra.keys import mask_to_names, DEFAULT_MODIFIER_MEANINGS, DEFAULT_MODIFIER_NUISANCE, DEFAULT_MODIFIER_IGNORE_KEYNAMES
+from xpra.client_base import XpraClientBase, EXIT_TIMEOUT
+from xpra.keys import DEFAULT_MODIFIER_MEANINGS, DEFAULT_MODIFIER_NUISANCE, DEFAULT_MODIFIER_IGNORE_KEYNAMES
 from xpra.platform.gui import ClientExtras
 from xpra.scripts.main import ENCODINGS
+from xpra.version_util import add_gtk_version_info
+from xpra.maths import std_unit
 
 from xpra.client_window import ClientWindow
 ClientWindowClass = ClientWindow
@@ -84,88 +92,67 @@ class XpraClient(XpraClientBase):
         self.start_time = time.time()
         self._window_to_id = {}
         self._id_to_window = {}
+        self._ui_events = 0
         self.title = opts.title
-        self.readonly = opts.readonly
         self.session_name = opts.session_name
-        self.compression_level = opts.compression_level
         self.auto_refresh_delay = opts.auto_refresh_delay
         self.max_bandwidth = opts.max_bandwidth
-        if self.max_bandwidth>0.0 and self.jpegquality==0:
-            """ jpegquality was not set, use a better start value """
-            self.jpegquality = 50
+        if self.max_bandwidth>0.0 and self.quality==0:
+            """ quality was not set, use a better start value """
+            self.quality = 80
+        self.dpi = int(opts.dpi)
 
-        self.mmap_enabled = False
+        #draw thread:
+        self._draw_queue = Queue()
+        self._draw_thread = Thread(target=self._draw_thread_loop, name="draw_loop")
+        self._draw_thread.setDaemon(True)
+        self._draw_thread.start()
+
+        #statistics and server info:
         self.server_start_time = -1
         self.server_platform = ""
         self.server_actual_desktop_size = None
+        self.server_max_desktop_size = None
+        self.server_display = None
         self.server_randr = False
         self.pixel_counter = maxdeque(maxlen=100)
-        self.server_latency = maxdeque(maxlen=100)
+        self.server_ping_latency = maxdeque(maxlen=100)
         self.server_load = None
-        self.client_latency = maxdeque(maxlen=100)
+        self.client_ping_latency = maxdeque(maxlen=100)
+        self.last_ping_echoed_time = 0
+
+        #features:
         self.toggle_cursors_bell_notify = False
         self.toggle_keyboard_sync = False
-        self.bell_enabled = True
-        self.cursors_enabled = True
-        self.notifications_enabled = True
-        self.clipboard_enabled = False
         self.window_configure = False
+        self.change_quality = False
+        self.readonly = opts.readonly
+        self.windows_enabled = opts.windows_enabled
+        self._client_extras = ClientExtras(self, opts, conn)
+        self.client_supports_notifications = opts.notifications and self._client_extras.can_notify()
+        self.client_supports_clipboard = opts.clipboard and self._client_extras.supports_clipboard() and not self.readonly
+        self.client_supports_cursors = opts.cursors
+        self.client_supports_bell = opts.bell
+        self.client_supports_sharing = opts.sharing
+        self.notifications_enabled = self.client_supports_notifications
+        self.clipboard_enabled = self.client_supports_clipboard
+        self.cursors_enabled = self.client_supports_cursors
+        self.bell_enabled = self.client_supports_bell
+
+        #mmap:
+        self.mmap_enabled = False
         self.mmap = None
         self.mmap_token = None
         self.mmap_file = None
         self.mmap_size = 0
-        self.last_ping_echoed_time = 0
-
-        self._client_extras = ClientExtras(self, opts)
-        self.clipboard_enabled = not self.readonly and opts.clipboard and self._client_extras.supports_clipboard()
         self.supports_mmap = opts.mmap and ("rgb24" in ENCODINGS) and self._client_extras.supports_mmap()
         if self.supports_mmap:
-            try:
-                import mmap
-                import tempfile
-                import uuid
-                from stat import S_IRUSR,S_IWUSR,S_IRGRP,S_IWGRP
-                mmap_dir = os.getenv("TMPDIR", "/tmp")
-                if not os.path.exists(mmap_dir):
-                    raise Exception("TMPDIR %s does not exist!" % mmap_dir)
-                #create the mmap file, the mkstemp that is called via NamedTemporaryFile ensures
-                #that the file is readable and writable only by the creating user ID
-                temp = tempfile.NamedTemporaryFile(prefix="xpra.", suffix=".mmap", dir=mmap_dir)
-                #keep a reference to it so it does not disappear!
-                self._mmap_temp_file = temp
-                self.mmap_file = temp.name
-                fd = temp.file.fileno()
-                #set the group permissions and gid if the mmap-group option is specified
-                if opts.mmap_group and type(conn.target)==str and os.path.exists(conn.target):
-                    s = os.stat(conn.target)
-                    os.fchown(fd, -1, s.st_gid)
-                    os.fchmod(fd, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP)
-                self.mmap_size = max(4096, mmap.PAGESIZE)*32*1024   #generally 128MB
-                log("using mmap file %s, fd=%s, size=%s", self.mmap_file, fd, self.mmap_size)
-                os.lseek(fd, self.mmap_size-1, os.SEEK_SET)
-                assert os.write(fd, '\x00')
-                os.lseek(fd, 0, os.SEEK_SET)
-                self.mmap = mmap.mmap(fd, length=self.mmap_size)
-                #write the 16 byte token one byte at a time - no endianness
-                self.mmap_token = uuid.uuid4().int
-                log.debug("mmap_token=%s", self.mmap_token)
-                v = self.mmap_token
-                for i in range(0,16):
-                    poke = ctypes.c_ubyte.from_buffer(self.mmap, 512+i)
-                    poke.value = v % 256
-                    v = v>>8
-                assert v==0
-            except Exception, e:
-                log.error("failed to setup mmap: %s", e)
-                self.supports_mmap = False
-                self.clean_mmap()
-                self.mmap = None
-                self.mmap_file = None
-                self.mmap_size = 0
+            self.init_mmap(opts.mmap_group, conn.filename)
 
         self.init_packet_handlers()
         self.ready(conn)
 
+        #keyboard:
         self.keyboard_sync = opts.keyboard_sync
         self.key_repeat_delay = -1
         self.key_repeat_interval = -1
@@ -182,22 +169,20 @@ class XpraClient(XpraClientBase):
 
         if self._keymap:
             self._keymap.connect("keys-changed", self._keys_changed)
-        self._xsettings_watcher = None
-        self._root_props_watcher = None
 
         self._focused = None
         def compute_receive_bandwidth(delay):
-            bytecount = self._protocol.input_bytecount
+            bytecount = conn.input_bytecount
             bw = ((bytecount - self.last_input_bytecount) / 1024) * 1000 / delay
             self.last_input_bytecount = bytecount;
             log.debug("Bandwidth is ", bw, "kB/s, max ", self.max_bandwidth, "kB/s")
-            q = self.jpegquality
+            q = self.quality
             if bw > self.max_bandwidth:
                 q -= 10
             elif bw < self.max_bandwidth:
                 q += 5
             q = max(10, min(95 ,q))
-            self.send_jpeg_quality(q)
+            self.send_quality(q)
             return True
         if (self.max_bandwidth):
             self.last_input_bytecount = 0
@@ -207,6 +192,51 @@ class XpraClient(XpraClientBase):
         else:
             gobject.timeout_add(20*1000, self.send_ping)
 
+    def init_mmap(self, mmap_group, socket_filename):
+        log("init_mmap(%s, %s)", mmap_group, socket_filename)
+        try:
+            import mmap
+            import tempfile
+            import uuid
+            from stat import S_IRUSR,S_IWUSR,S_IRGRP,S_IWGRP
+            mmap_dir = os.getenv("TMPDIR", "/tmp")
+            if not os.path.exists(mmap_dir):
+                raise Exception("TMPDIR %s does not exist!" % mmap_dir)
+            #create the mmap file, the mkstemp that is called via NamedTemporaryFile ensures
+            #that the file is readable and writable only by the creating user ID
+            temp = tempfile.NamedTemporaryFile(prefix="xpra.", suffix=".mmap", dir=mmap_dir)
+            #keep a reference to it so it does not disappear!
+            self._mmap_temp_file = temp
+            self.mmap_file = temp.name
+            fd = temp.file.fileno()
+            #set the group permissions and gid if the mmap-group option is specified
+            if mmap_group and type(socket_filename)==str and os.path.exists(socket_filename):
+                s = os.stat(socket_filename)
+                os.fchown(fd, -1, s.st_gid)
+                os.fchmod(fd, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP)
+            self.mmap_size = max(4096, mmap.PAGESIZE)*32*1024   #generally 128MB
+            log("using mmap file %s, fd=%s, size=%s", self.mmap_file, fd, self.mmap_size)
+            os.lseek(fd, self.mmap_size-1, os.SEEK_SET)
+            assert os.write(fd, '\x00')
+            os.lseek(fd, 0, os.SEEK_SET)
+            self.mmap = mmap.mmap(fd, length=self.mmap_size)
+            #write the 16 byte token one byte at a time - no endianness
+            self.mmap_token = uuid.uuid4().int
+            log.debug("mmap_token=%s", self.mmap_token)
+            v = self.mmap_token
+            for i in range(0,16):
+                poke = ctypes.c_ubyte.from_buffer(self.mmap, 512+i)
+                poke.value = v % 256
+                v = v>>8
+            assert v==0
+        except Exception, e:
+            log.error("failed to setup mmap: %s", e)
+            self.supports_mmap = False
+            self.clean_mmap()
+            self.mmap = None
+            self.mmap_file = None
+            self.mmap_size = 0
+
     def init_packet_handlers(self):
         XpraClientBase.init_packet_handlers(self)
         for k,v in {
@@ -214,19 +244,23 @@ class XpraClient(XpraClientBase):
             "new-window":           self._process_new_window,
             "new-override-redirect":self._process_new_override_redirect,
             "window-resized":       self._process_window_resized,
-            "draw":                 self._process_draw,
             "cursor":               self._process_cursor,
             "bell":                 self._process_bell,
             "notify_show":          self._process_notify_show,
             "notify_close":         self._process_notify_close,
-            "ping":                 self._process_ping,
-            "ping_echo":            self._process_ping_echo,
             "window-metadata":      self._process_window_metadata,
             "configure-override-redirect":  self._process_configure_override_redirect,
             "lost-window":          self._process_lost_window,
             "desktop_size":         self._process_desktop_size,
             # "clipboard-*" packets are handled by a special case below.
             }.items():
+            self._ui_packet_handlers[k] = v
+        #these handlers can run directly from the network thread:
+        for k,v in {
+            "draw":                 self._process_draw,
+            "ping":                 self._process_ping,
+            "ping_echo":            self._process_ping_echo,
+            }.items():
             self._packet_handlers[k] = v
 
     def run(self):
@@ -242,8 +276,7 @@ class XpraClient(XpraClientBase):
 
     def cleanup(self):
         if self._client_extras:
-            self._client_extras.exit()
-            self._client_extras = None
+            self._client_extras.cleanup()
         XpraClientBase.cleanup(self)
         self.clean_mmap()
 
@@ -333,31 +366,21 @@ class XpraClient(XpraClientBase):
             log.error("key_handled_as_shortcut(%s,%s,%s,%s) failed to execute shortcut=%s: %s", window, key_name, modifiers, depressed, shortcut, e)
         return  True
 
-    def handle_key_action(self, event, window, depressed):
+    def handle_key_action(self, event, window, pressed):
         if self.readonly:
             return
-        log.debug("handle_key_action(%s,%s,%s)", event, window, depressed)
-        modifiers = self.mask_to_names(event.state)
-        name = gdk.keyval_name(event.keyval)
-        keyval = nn(event.keyval)
-        keycode = event.hardware_keycode
-        group = event.group
-        #meant to be in PyGTK since 2.10, not used yet so just return False if we don't have it:
-        is_modifier = hasattr(event, "is_modifier") and event.is_modifier
-        translated = self._client_extras.translate_key(depressed, keyval, name, keycode, group, is_modifier, modifiers)
-        if translated is None:
-            return
-        depressed, keyval, name, keycode, group, is_modifier, modifiers = translated
-        if self.key_handled_as_shortcut(window, name, modifiers, depressed):
-            return
-        if keycode<0:
-            log.debug("key_action(%s,%s,%s) translated keycode is %s, ignoring it", event, window, depressed, keycode)
-            return
-        log.debug("key_action(%s,%s,%s) modifiers=%s, name=%s, state=%s, keyval=%s, string=%s, keycode=%s", event, window, depressed, modifiers, name, event.state, event.keyval, event.string, keycode)
+        #NOTE: handle_key_event may fire send_key_action more than once (see win32 AltGr)
         wid = self._window_to_id[window]
-        self.send(["key-action", wid, nn(name), depressed, modifiers, keyval, nn(event.string), nn(keycode), group, is_modifier])
+        self._client_extras.handle_key_event(self.send_key_action, event, wid, pressed)
+
+    def send_key_action(self, wid, keyname, pressed, modifiers, keyval, string, keycode, group, is_modifier):
+        window = self._id_to_window[wid]
+        if self.key_handled_as_shortcut(window, keyname, modifiers, pressed):
+            return
+        log("send_key_action(%s, %s, %s, %s, %s, %s, %s, %s, %s)", wid, keyname, pressed, modifiers, keyval, string, keycode, group, is_modifier)
+        self.send(["key-action", wid, nn(keyname), pressed, modifiers, nn(keyval), string, nn(keycode), group, is_modifier])
         if self.keyboard_sync and self.key_repeat_delay>0 and self.key_repeat_interval>0:
-            self._key_repeat(wid, depressed, name, keyval, keycode)
+            self._key_repeat(wid, pressed, keyname, keyval, keycode)
 
     def _key_repeat(self, wid, depressed, name, keyval, keycode):
         """ this method takes care of scheduling the sending of
@@ -365,7 +388,7 @@ class XpraClient(XpraClientBase):
             maintain a consistent keyboard state.
         """
         #we keep track of which keys are still pressed in a dict,
-        if keycode==0:
+        if keycode<0:
             key = name
         else:
             key = keycode
@@ -421,10 +444,12 @@ class XpraClient(XpraClientBase):
             self.xkbmap_layout, self.xkbmap_variant, self.xkbmap_variants = self._client_extras.get_layout_spec()
             self.xkbmap_print, self.xkbmap_query = self._client_extras.get_keymap_spec()
         self.xkbmap_keycodes = self._client_extras.get_gtk_keymap()
+        self.xkbmap_x11_keycodes = self._client_extras.get_x11_keymap()
         self.xkbmap_mod_meanings, self.xkbmap_mod_managed, self.xkbmap_mod_pointermissing = self._client_extras.get_keymap_modifiers()
         log.debug("layout=%s, variant=%s", self.xkbmap_layout, self.xkbmap_variant)
         log.debug("print=%s, query=%s", self.xkbmap_print, self.xkbmap_query)
         log.debug("keycodes=%s", str(self.xkbmap_keycodes)[:80]+"...")
+        log.debug("x11 keycodes=%s", str(self.xkbmap_x11_keycodes)[:80]+"...")
         log.debug("xkbmap_mod_meanings: %s", self.xkbmap_mod_meanings)
 
     def _keys_changed(self, *args):
@@ -456,7 +481,7 @@ class XpraClient(XpraClientBase):
     def get_keymap_properties(self):
         props = {"modifiers" : self.get_current_modifiers()}
         for x in ["xkbmap_print", "xkbmap_query", "xkbmap_mod_meanings",
-              "xkbmap_mod_managed", "xkbmap_mod_pointermissing", "xkbmap_keycodes"]:
+              "xkbmap_mod_managed", "xkbmap_mod_pointermissing", "xkbmap_keycodes", "xkbmap_x11_keycodes"]:
             props[x] = nn(getattr(self, x))
         return  props
 
@@ -479,20 +504,27 @@ class XpraClient(XpraClientBase):
         return self.mask_to_names(modifiers_mask)
 
     def mask_to_names(self, mask):
-        mn = mask_to_names(mask, self._modifier_map)
-        names = mn
-        if self._client_extras:
-            names = self._client_extras.current_modifiers(mn)
-        return  names
+        if self._client_extras is None:
+            return []
+        return self._client_extras.mask_to_names(mask)
 
     def send_positional(self, packet):
-        self._protocol.source.queue_positional_packet(packet)
+        p = self._protocol
+        if p is not None:
+            s = p.source
+            if s:
+                s.queue_positional_packet(packet)
 
     def send_mouse_position(self, packet):
-        self._protocol.source.queue_mouse_position_packet(packet)
+        p = self._protocol
+        if p is not None:
+            s = p.source
+            if s:
+                s.queue_mouse_position_packet(packet)
 
     def make_hello(self, challenge_response=None):
         capabilities = XpraClientBase.make_hello(self, challenge_response)
+        add_gtk_version_info(capabilities, gtk)
         for k,v in self.get_keymap_properties().items():
             capabilities[k] = v
         if self.readonly:
@@ -501,11 +533,11 @@ class XpraClient(XpraClientBase):
         else:
             capabilities["xkbmap_layout"] = nn(self.xkbmap_layout)
             capabilities["xkbmap_variant"] = nn(self.xkbmap_variant)
-        capabilities["clipboard"] = self.clipboard_enabled
-        capabilities["notifications"] = self._client_extras.can_notify()
         capabilities["modifiers"] = self.get_current_modifiers()
         root_w, root_h = get_root_size()
         capabilities["desktop_size"] = [root_w, root_h]
+        capabilities["screen_sizes"] = self.get_screen_sizes()
+        capabilities["client_type"] = "Python/Gtk"
         key_repeat = self._client_extras.get_keyboard_repeat()
         if key_repeat:
             delay_ms,interval_ms = key_repeat
@@ -515,33 +547,44 @@ class XpraClient(XpraClientBase):
             capabilities["mmap_file"] = self.mmap_file
             capabilities["mmap_token"] = self.mmap_token
         capabilities["randr_notify"] = True
-        #these should be turned into options:
-        capabilities["cursors"] = True
-        capabilities["bell"] = True
-        capabilities["png_window_icons"] = "png" in ENCODINGS
+        capabilities["compressible_cursors"] = True
+        capabilities["dpi"] = self.dpi
+        capabilities["clipboard"] = self.client_supports_clipboard
+        capabilities["notifications"] = self.client_supports_notifications
+        capabilities["cursors"] = self.client_supports_cursors
+        capabilities["bell"] = self.client_supports_bell
+        capabilities["encoding_client_options"] = True
+        capabilities["rgb24zlib"] = True
+        capabilities["share"] = self.client_supports_sharing
+        capabilities["auto_refresh_delay"] = int(self.auto_refresh_delay*1000)
+        capabilities["windows"] = self.windows_enabled
+        try:
+            from wimpiggy.prop import set_xsettings_format
+            assert set_xsettings_format
+            capabilities["xsettings-tuple"] = True
+        except:
+            pass
         return capabilities
 
     def send_ping(self):
         now_ms = int(1000*time.time())
         self.send(["ping", now_ms])
-        wait = 15
+        wait = 60
         def check_echo_received(*args):
             if self.last_ping_echoed_time<now_ms:
-                log.error("check_echo_received: we sent a ping to the server %s seconds ago and we have not received its echo!", wait)
-                log.error("    assuming that the connection is dead and disconnecting")
-                self.quit(1)
+                self.warn_and_quit(EXIT_TIMEOUT, "server ping timeout - waited %s seconds without a response" % wait)
         gobject.timeout_add(wait*1000, check_echo_received)
         return True
 
     def _process_ping_echo(self, packet):
         echoedtime, l1, l2, l3, cl = packet[1:6]
         self.last_ping_echoed_time = echoedtime
-        diff = int(1000*time.time()-echoedtime)
-        self.server_latency.append(diff)
-        self.server_load = (l1, l2, l3)
+        server_ping_latency = time.time()-echoedtime/1000.0
+        self.server_ping_latency.append((time.time(), server_ping_latency))
+        self.server_load = l1, l2, l3
         if cl>=0:
-            self.client_latency.append(cl)
-        log("ping echo server load=%s, measured client latency=%s", self.server_load, cl)
+            self.client_ping_latency.append((time.time(), cl/1000.0))
+        log("ping echo server load=%s, measured client latency=%sms", self.server_load, cl)
 
     def _process_ping(self, packet):
         echotime = packet[1]
@@ -551,25 +594,28 @@ class XpraClient(XpraClientBase):
         except:
             l1,l2,l3 = 0,0,0
         sl = -1
-        if len(self.server_latency)>0:
-            sl = self.server_latency[-1]
-        self.send(["ping_echo", echotime, l1, l2, l3, sl])
+        if len(self.server_ping_latency)>0:
+            _, sl = self.server_ping_latency[-1]
+        self.idle_send(["ping_echo", echotime, l1, l2, l3, int(1000.0*sl)])
 
-    def send_jpeg_quality(self, q):
-        assert q>0 and q<100
-        self.jpegquality = q
-        self.send(["jpeg-quality", self.jpegquality])
+    def send_quality(self, q):
+        assert q==-1 or (q>=0 and q<=100), "invalid quality: %s" % q
+        self.quality = q
+        if self.change_quality:
+            self.send(["quality", self.quality])
+        else:
+            self.send(["jpeg-quality", self.quality])
 
     def send_refresh(self, wid):
         self.send(["buffer-refresh", wid, True, 95])
-        self._refresh_requested = True
 
     def send_refresh_all(self):
         log.debug("Automatic refresh for all windows ")
         self.send_refresh(-1)
 
     def parse_server_capabilities(self, capabilities):
-        XpraClientBase.parse_server_capabilities(self, capabilities)
+        if not XpraClientBase.parse_server_capabilities(self, capabilities):
+            return
         if not self.session_name:
             self.session_name = capabilities.get("session_name", "Xpra")
         try:
@@ -583,6 +629,7 @@ class XpraClient(XpraClientBase):
         log("server actual desktop size=%s", self.server_actual_desktop_size)
         self.set_max_packet_size()
         self.server_max_desktop_size = capabilities.get("max_desktop_size")
+        self.server_display = capabilities.get("display")
         server_desktop_size = capabilities.get("desktop_size")
         log("server desktop size=%s", server_desktop_size)
         assert server_desktop_size
@@ -609,12 +656,25 @@ class XpraClient(XpraClientBase):
             log.debug("server is using %s encoding" % e)
             self.encoding = e
         self.window_configure = capabilities.get("window_configure", False)
-        self.notifications_enabled = capabilities.get("notifications", False)
-        clipboard_server_support = capabilities.get("clipboard", True)
-        self.clipboard_enabled = clipboard_server_support and self._client_extras.supports_clipboard()
+        self.server_supports_notifications = capabilities.get("notifications", False)
+        self.notifications_enabled = self.server_supports_notifications and self.client_supports_notifications
+        self.server_supports_cursors = capabilities.get("cursors", True)    #added in 0.5, default to True!
+        self.cursors_enabled = self.server_supports_cursors and self.client_supports_cursors
+        try:
+            from wimpiggy.prop import set_xsettings_format
+            set_xsettings_format(use_tuple=capabilities.get("xsettings-tuple", False))
+        except Exception, e:
+            if os.name=="posix" and not sys.platform.startswith("darwin"):
+                log.error("failed to set xsettings format: %s", e)
+        self.server_supports_bell = capabilities.get("bell", True)          #added in 0.5, default to True!
+        self.bell_enabled = self.server_supports_bell and self.client_supports_bell
+        self.server_supports_clipboard = capabilities.get("clipboard", False)
+        self.clipboard_enabled = self.client_supports_clipboard and self.server_supports_clipboard
         self.mmap_enabled = self.supports_mmap and self.mmap_file and capabilities.get("mmap_enabled")
+        self.server_auto_refresh_delay = capabilities.get("auto_refresh_delay", 0)/1000
+        self.change_quality = capabilities.get("change-quality", False)
         if self.mmap_enabled:
-            log.info("mmap enabled using %s", self.mmap_file)
+            log.info("mmap is enabled using %sBytes area in %s", std_unit(self.mmap_size), self.mmap_file)
         #the server will have a handle on the mmap file by now, safe to delete:
         self.clean_mmap()
         self.send_deflate_level()
@@ -622,29 +682,45 @@ class XpraClient(XpraClientBase):
         self.server_platform = capabilities.get("platform")
         self.toggle_cursors_bell_notify = capabilities.get("toggle_cursors_bell_notify", False)
         self.toggle_keyboard_sync = capabilities.get("toggle_keyboard_sync", False)
+        modifier_keycodes = capabilities.get("modifier_keycodes")
+        if modifier_keycodes:
+            self._client_extras.set_modifier_mappings(modifier_keycodes)
+
         #ui may want to know this is now set:
         self.emit("clipboard-toggled")
         self.key_repeat_delay, self.key_repeat_interval = capabilities.get("key_repeat", (-1,-1))
         self.emit("handshake-complete")
-        if clipboard_server_support:
+        if self.clipboard_enabled:
             #from now on, we will send a message to the server whenever the clipboard flag changes:
             self.connect("clipboard-toggled", self.send_clipboard_enabled_status)
         if self.toggle_keyboard_sync:
             self.connect("keyboard-sync-toggled", self.send_keyboard_sync_enabled_status)
+        self.send_ping()
 
     def send_notify_enabled(self):
-        if self.toggle_cursors_bell_notify:
-            self.send(["set-notify", self.notifications_enabled])
+        assert self.client_supports_notifications, "cannot toggle notifications: the feature is disabled by the client"
+        assert self.server_supports_notifications, "cannot toggle notifications: the feature is disabled by the server"
+        assert self.toggle_cursors_bell_notify, "cannot toggle notifications: server lacks the feature"
+        self.send(["set-notify", self.notifications_enabled])
 
     def send_bell_enabled(self):
-        if self.toggle_cursors_bell_notify:
-            self.send(["set-bell", self.bell_enabled])
+        assert self.client_supports_bell, "cannot toggle bell: the feature is disabled by the client"
+        assert self.server_supports_bell, "cannot toggle bell: the feature is disabled by the server"
+        assert self.toggle_cursors_bell_notify, "cannot toggle bell: server lacks the feature"
+        self.send(["set-bell", self.bell_enabled])
 
     def send_cursors_enabled(self):
-        if self.toggle_cursors_bell_notify:
-            self.send(["set-cursors", self.cursors_enabled])
+        assert self.client_supports_cursors, "cannot toggle cursors: the feature is disabled by the client"
+        assert self.server_supports_cursors, "cannot toggle cursors: the feature is disabled by the server"
+        assert self.toggle_cursors_bell_notify, "cannot toggle cursors: server lacks the feature"
+        self.send(["set-cursors", self.cursors_enabled])
+
+    def set_deflate_level(self, level):
+        self.compression_level = level
+        self.send_deflate_level()
 
     def send_deflate_level(self):
+        self._protocol.set_compression_level(self.compression_level)
         self.send(["set_deflate", self.compression_level])
 
     def send_clipboard_enabled_status(self, *args):
@@ -663,17 +739,52 @@ class XpraClient(XpraClientBase):
     def _screen_size_changed(self, *args):
         root_w, root_h = get_root_size()
         log.debug("sending updated screen size to server: %sx%s", root_w, root_h)
-        self.send(["desktop_size", root_w, root_h])
+        self.send(["desktop_size", root_w, root_h, self.get_screen_sizes()])
         #update the max packet size (may have gone up):
         self.set_max_packet_size()
 
+    def get_screen_sizes(self):
+        screen_sizes = []
+        display = gdk.display_get_default()
+        i=0
+        while i<display.get_n_screens():
+            screen = display.get_screen(i)
+            j = 0
+            monitors = []
+            while j<screen.get_n_monitors():
+                geom = screen.get_monitor_geometry(j)
+                monitor = (screen.get_monitor_plug_name(j) or "", geom.x, geom.y, geom.width, geom.height,
+                            screen.get_monitor_width_mm(j), screen.get_monitor_height_mm(j))
+                monitors.append(monitor)
+                j += 1
+            item = (screen.make_display_name(), screen.get_width(), screen.get_height(),
+                        screen.get_width_mm(), screen.get_height_mm(),
+                        monitors)
+            screen_sizes.append(item)
+            i += 1
+        return screen_sizes
+
+    def _ui_event(self):
+        if self._ui_events==0:
+            self.emit("first-ui-received")
+        self._ui_events += 1
+
     def _process_new_common(self, packet, override_redirect):
-        (wid, x, y, w, h, metadata) = packet[1:7]
+        self._ui_event()
+        wid, x, y, w, h, metadata = packet[1:7]
+        assert wid not in self._id_to_window, "we already have a window %s" % wid
         if w<=0 or h<=0:
             log.error("window dimensions are wrong: %sx%s", w, h)
             w = 10
             h = 5
-        window = ClientWindowClass(self, wid, x, y, w, h, metadata, override_redirect)
+        client_properties = {}
+        if len(packet)>=8:
+            client_properties = packet[7]
+        if self.server_auto_refresh_delay>0:
+            auto_refresh_delay = 0                          #server takes care of it
+        else:
+            auto_refresh_delay = self.auto_refresh_delay    #we do it
+        window = ClientWindowClass(self, wid, x, y, w, h, metadata, override_redirect, client_properties, auto_refresh_delay)
         self._id_to_window[wid] = window
         self._window_to_id[window] = wid
         window.show_all()
@@ -692,27 +803,72 @@ class XpraClient(XpraClientBase):
             window.resize(w, h)
 
     def _process_draw(self, packet):
-        (wid, x, y, width, height, coding, data, packet_sequence, rowstride) = packet[1:10]
-        window = self._id_to_window.get(wid)
-        decode_time = 0
-        if window:
-            start = time.time()
-            if window.draw_region(x, y, width, height, coding, data, rowstride):
-                end = time.time()
-                self.pixel_counter.append((end, width*height))
-                decode_time = int(end*1000*1000-start*1000*1000)
-        else:
-            #window is gone
-            if coding=="mmap":
-                #we need to ack the data to free the space!
-                assert self.mmap_enabled
-                data_start = ctypes.c_uint.from_buffer(self.mmap, 0)
-                offset, length = data[-1]
-                data_start.value = offset+length
+        self._draw_queue.put(packet)
+
+    def send_damage_sequence(self, wid, packet_sequence, width, height, decode_time):
         self.send_now(["damage-sequence", packet_sequence, wid, width, height, decode_time])
 
+    def _draw_thread_loop(self):
+        while self.exit_code is None:
+            packet = self._draw_queue.get()
+            try:
+                self._do_draw(packet)
+                time.sleep(0)
+            except:
+                log.error("error processing draw packet", exc_info=True)
+
+    def _do_draw(self, packet):
+        """ this runs from the draw thread above """
+        wid, x, y, width, height, coding, data, packet_sequence, rowstride = packet[1:10]
+        window = self._id_to_window.get(wid)
+        if not window:
+            #window is gone
+            if coding=="mmap":
+                assert self.mmap_enabled
+                def free_mmap_area():
+                    #we need to ack the data to free the space!
+                    data_start = ctypes.c_uint.from_buffer(self.mmap, 0)
+                    offset, length = data[-1]
+                    data_start.value = offset+length
+                #clear the mmap area via idle_add so any pending draw requests
+                #will get a chance to run first
+                gobject.idle_add(free_mmap_area)
+            self.send_damage_sequence(wid, packet_sequence, width, height, -1)
+            return
+        options = {}
+        if len(packet)>10:
+            options = packet[10]
+        log("process_draw %s bytes for window %s using %s encoding with options=%s", len(data), wid, coding, options)
+        start = time.time()
+        def record_decode_time(success):
+            if success:
+                end = time.time()
+                decode_time = int(end*1000*1000-start*1000*1000)
+                self.pixel_counter.append((end, width*height))
+            else:
+                decode_time = -1
+            log("record_decode_time(%s) wid=%s, %s: %sx%s, %sms", success, wid, coding, width, height, int(decode_time/100)/10.0)
+            self.send_damage_sequence(wid, packet_sequence, width, height, decode_time)
+        window.draw_region(x, y, width, height, coding, data, rowstride, packet_sequence, options, [record_decode_time])
+
     def _process_cursor(self, packet):
-        (_, new_cursor) = packet
+        if not self.cursors_enabled:
+            return
+        if len(packet)==2:
+            new_cursor = packet[1]
+        elif len(packet)>=8:
+            new_cursor = packet[1:]
+        else:
+            raise Exception("invalid cursor packet: %s items" % len(packet))
+        if len(new_cursor)>0:
+            pixels = new_cursor[7]
+            if type(pixels)==tuple:
+                #newer versions encode as a list, see "compressible_cursors" capability
+                import array
+                a = array.array('b', '\0'* len(pixels))
+                a.fromlist(list(pixels))
+                new_cursor = list(new_cursor)
+                new_cursor[7] = a
         set_windows_cursor(self._id_to_window.values(), new_cursor)
 
     def _process_bell(self, packet):
@@ -733,7 +889,8 @@ class XpraClient(XpraClientBase):
     def _process_notify_show(self, packet):
         if not self.notifications_enabled:
             return
-        (dbus_id, nid, app_name, replaces_nid, app_icon, summary, body, expire_timeout) = packet[1:9]
+        self._ui_event()
+        dbus_id, nid, app_name, replaces_nid, app_icon, summary, body, expire_timeout = packet[1:9]
         log("_process_notify_show(%s)", packet)
         self._client_extras.show_notify(dbus_id, nid, app_name, replaces_nid, app_icon, summary, body, expire_timeout)
 
@@ -792,7 +949,7 @@ class XpraClient(XpraClientBase):
         packet_type = str(packet[0])
         if packet_type.startswith("clipboard-"):
             if self.clipboard_enabled:
-                self._client_extras.process_clipboard_packet(packet)
+                gobject.idle_add(self._client_extras.process_clipboard_packet, packet)
         else:
             XpraClientBase.process_packet(self, proto, packet)
 
