@@ -5,6 +5,9 @@
 # Parti is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
+#disabled for now as this causes bugs:
+PRESERVE_WORSPACE = False
+
 #pygtk3 vs pygtk2 (sigh)
 from wimpiggy.gobject_compat import import_gobject, import_gtk, import_gdk, is_gtk3
 gobject = import_gobject()
@@ -15,6 +18,8 @@ if is_gtk3():
         #TODO: no idea how to do this with gtk3
         #maybe not even possible..
         gtk.Window.__init__(win)
+    def is_mapped(win):
+        return win.get_mapped()
     def get_window_geometry(gtkwindow):
         x, y = gtkwindow.get_position()
         w, h = gtkwindow.get_size()
@@ -48,7 +53,7 @@ if is_gtk3():
                          }
         geom = gdk.Geometry()
         mask = 0
-        for k,v in hints:
+        for k,v in hints.items():
             if k in INT_FIELDS:
                 setattr(geom, k, int(v))
                 mask |= int(name_to_hint.get(k, 0))
@@ -70,6 +75,8 @@ if is_gtk3():
 else:
     def init_window(gtkwindow, wintype):
         gtk.Window.__init__(gtkwindow, wintype)
+    def is_mapped(win):
+        return win.window is not None and win.window.is_visible()
     def get_window_geometry(gtkwindow):
         gdkwindow = gtkwindow.get_window()
         x, y = gdkwindow.get_origin()
@@ -79,7 +86,12 @@ else:
         gtkwindow.set_geometry_hints(None, **hints)
 
     def queue_draw(gtkwindow, x, y, width, height):
-        gtkwindow.get_window().invalidate_rect(gdk.Rectangle(x, y, width, height), False)
+        window = gtkwindow.get_window()
+        if window:
+            window.invalidate_rect(gdk.Rectangle(x, y, width, height), False)
+        else:
+            log.warn("ignoring draw received for a window which is not realized yet!")
+
     WINDOW_POPUP = gtk.WINDOW_POPUP
     WINDOW_TOPLEVEL = gtk.WINDOW_TOPLEVEL
     WINDOW_EVENT_MASK = gdk.STRUCTURE_MASK | gdk.KEY_PRESS_MASK | gdk.KEY_RELEASE_MASK | gdk.POINTER_MOTION_MASK | gdk.BUTTON_PRESS_MASK | gdk.BUTTON_RELEASE_MASK
@@ -124,6 +136,11 @@ from wimpiggy.log import Logger
 log = Logger()
 
 from xpra.window_backing import new_backing
+try:
+    from wimpiggy.prop import prop_set, prop_get
+    has_wimpiggy_prop = True
+except ImportError, e:
+    has_wimpiggy_prop = False
 
 if sys.version < '3':
     import codecs
@@ -136,7 +153,7 @@ else:
 
 
 class ClientWindow(gtk.Window):
-    def __init__(self, client, wid, x, y, w, h, metadata, override_redirect):
+    def __init__(self, client, wid, x, y, w, h, metadata, override_redirect, client_properties, auto_refresh_delay):
         if override_redirect:
             init_window(self, WINDOW_POPUP)
         else:
@@ -149,8 +166,11 @@ class ClientWindow(gtk.Window):
         self.new_backing(w, h)
         self._metadata = {}
         self._override_redirect = override_redirect
+        self._client_properties = client_properties
+        self._auto_refresh_delay = auto_refresh_delay
         self._refresh_timer = None
-        self._refresh_requested = False
+        self._refresh_min_pixels = -1
+        self._refresh_ignore_sequence = -1
         # used for only sending focus events *after* the window is mapped:
         self._been_mapped = False
         self._override_redirect_windows = []
@@ -171,6 +191,105 @@ class ClientWindow(gtk.Window):
                 transient_for._override_redirect_windows.append(self)
         self.connect("notify::has-toplevel-focus", self._focus_change)
 
+    def do_realize(self):
+        if not PRESERVE_WORSPACE:
+            gtk.Window.do_realize(self)
+            return
+        ndesktops = 0
+        try:
+            root = gtk.gdk.screen_get_default().get_root_window()
+            prop = root.property_get("_NET_NUMBER_OF_DESKTOPS")
+            if prop is not None and len(prop)==3 and len(prop[2])==1:
+                ndesktops = prop[2][0]
+        except Exception, e:
+            log.error("failed to get workspace count: %s", e)
+        workspace = self._client_properties.get("workspace", -1)
+        log("do_realize() workspace=%s (ndesktops=%s)", workspace, ndesktops)
+
+        #below we duplicate gtk.window.realize() code
+        #just so we can insert the property code at the right place:
+        #after the gdk.Window is created, but before it gets positionned.
+        allocation = self.get_allocation()
+        if allocation.x==-1 and allocation.y==-1 and allocation.width==1 and allocation.height==1:
+            w, h = self.size_request()
+            if w>0 or h>0:
+                allocation.width = w
+                allocation.height = h
+            self.size_allocate(w, h)
+            self.queue_resize()
+            if self.flags() & gtk.REALIZED:
+                log.error("window is already realized!")
+                return
+
+        self.set_flags(gtk.REALIZED)
+        is_toplevel = self.get_parent() is None
+        if hasattr(self, "is_toplevel"):
+            is_toplevel = self.is_toplevel()
+        if is_toplevel:
+            window_type = gtk.gdk.WINDOW_TOPLEVEL
+        else:
+            window_type = gtk.gdk.WINDOW_TEMP
+        if self.get_has_frame():
+            #TODO: duplicate gtk code here too..
+            pass
+
+        events = self.get_events() | gdk.EXPOSURE_MASK | gdk.STRUCTURE_MASK | \
+                    gdk.KEY_PRESS_MASK | gdk.KEY_RELEASE_MASK
+        self.window = gdk.Window(
+            self.get_root_window(),
+            x=allocation.x, y=allocation.y, width=allocation.width, height=allocation.height,
+            wmclass_name=self.wmclass_name, wmclass_class=self.wmclass_class,
+            window_type=window_type,
+            wclass=gdk.INPUT_OUTPUT,
+            title=self.get_title(),
+            event_mask=events,
+            )
+
+        if has_wimpiggy_prop and not self._override_redirect and ndesktops>workspace and workspace>=0:
+            try:
+                prop_set(self.window, "_NET_WM_DESKTOP", "u32", workspace)
+            except Exception, e:
+                log.error("failed to set workspace: %s", e)
+
+        self.window.set_opacity(1.0)
+        #self.window.enable_synchronized_configure().. not used?
+        self.window.set_user_data(self)
+        self.style.attach(self.window)
+        self.style.set_background(self.window, gtk.STATE_NORMAL)
+        #self.paint() does not exist in pygtk..
+        transient_for = self.get_transient_for()
+        if transient_for and transient_for.flags() & gtk.REALIZED:
+            self.window.set_transient_for(transient_for.get_window())
+        if not self.get_decorated():
+            self.window.set_decorations(0)
+        if not self.get_deletable():
+            self.window.set_functions(gtk.gdk.FUNC_ALL | gtk.gdk.FUNC_CLOSE)
+        if self.get_skip_pager_hint():
+            self.window.set_skip_pager_hint(True)
+        if self.get_skip_taskbar_hint():
+            self.window.set_skip_taskbar_hint(True)
+        self.window.set_accept_focus(self.get_accept_focus())
+        self.window.set_focus_on_map(self.get_focus_on_map())
+        self.window.set_modal_hint(self.get_modal())
+        #cannot access startup id...
+
+    def get_workspace(self):
+        try:
+            if sys.platform.startswith("win"):
+                return  -1              #windows does not have workspaces
+            if not has_wimpiggy_prop:
+                prop = self.window.get_screen().get_root_window().property_get("_NET_CURRENT_DESKTOP")
+                if not prop or len(prop)!=3 or len(prop[2])!=1:
+                    return  -1
+                return prop[2][0]
+            v = prop_get(self.get_window(), "_NET_WM_DESKTOP", "u32", ignore_errors=True)
+            if type(v)==int:
+                return  v
+        except Exception, e:
+            log.error("failed to detect workspace: %s", e)
+        return  -1
+
+
     def new_backing(self, w, h):
         self._backing = new_backing(self._id, w, h, self._backing, self._client.supports_mmap, self._client.mmap)
 
@@ -187,7 +306,9 @@ class ClientWindow(gtk.Window):
                 var = atvar[1:len(atvar)-1]     #ie: 'title'
                 default_value = default_values.get(var, u("<unknown %s>") % var)
                 value = self._metadata.get(var, default_value)
-                return value.decode("utf-8")
+                if sys.version<'3':
+                    value = value.decode("utf-8")
+                return value
             title = re.sub("@[\w\-]*@", metadata_replace, title)
         self.set_title(title)
 
@@ -200,15 +321,19 @@ class ClientWindow(gtk.Window):
                 ("base-size", "base_width", "base_height"),
                 ("increment", "width_inc", "height_inc"),
                 ]:
-                if a in self._metadata["size-constraints"]:
-                    hints[h1], hints[h2] = size_metadata[a]
+                v = size_metadata.get(a)
+                if v:
+                    hints[h1], hints[h2] = v
             for (a, h) in [
-                ("minimum-aspect", "min_aspect_ratio"),
-                ("maximum-aspect", "max_aspect_ratio"),
+                ("minimum-aspect", "min_aspect"),
+                ("maximum-aspect", "max_aspect"),
                 ]:
-                if a in self._metadata:
-                    hints[h] = size_metadata[a][0] * 1.0 / size_metadata[a][1]
+                v = size_metadata.get(a)
+                if v:
+                    hints[h] = v
             set_geometry_hints(self, hints)
+            #TODO:
+            #gravity = size_metadata.get("gravity")
 
         if hasattr(self, "get_realized"):
             #pygtk 2.22 and above have this method:
@@ -247,7 +372,8 @@ class ClientWindow(gtk.Window):
             if window:
                 self.set_transient_for(window)
 
-        if "window-type" in self._metadata:
+        #apply window-type hint if window is not mapped yet:
+        if "window-type" in self._metadata and not is_mapped(self):
             window_types = self._metadata.get("window-type")
             log("window types=%s", window_types)
             for window_type in window_types:
@@ -257,8 +383,8 @@ class ClientWindow(gtk.Window):
                     self.set_type_hint(hint)
                     break
 
-    def refresh_window(self):
-        log("Automatic refresh for id %s", self._id)
+    def refresh_window(self, *args):
+        log("refresh_window(%s) wid=%s", args, self._id)
         self._client.send_refresh(self._id)
 
     def refresh_all_windows(self):
@@ -267,19 +393,29 @@ class ClientWindow(gtk.Window):
         #the "ClientWindow"
         self._client.send_refresh_all()
 
-    def draw_region(self, x, y, width, height, coding, img_data, rowstride):
-        success = self._backing.draw_region(x, y, width, height, coding, img_data, rowstride)
-        if success:
-            queue_draw(self, x, y, width, height)
-        if self._refresh_requested:
-            self._refresh_requested = False
-        else:
-            if self._refresh_timer:
+    def draw_region(self, x, y, width, height, coding, img_data, rowstride, packet_sequence, options, callbacks):
+        def after_draw_refresh(success):
+            log("after_draw_refresh(%s) options=%s", success, options)
+            if success:
+                queue_draw(self, x, y, width, height)
+            #clear the auto refresh if enough pixels were sent (arbitrary limit..)
+            if success and self._refresh_timer and width*height>=self._refresh_min_pixels:
                 gobject.source_remove(self._refresh_timer)
                 self._refresh_timer = None
-            if self._client.auto_refresh_delay and coding == "jpeg":
-                self._refresh_timer = gobject.timeout_add(int(1000 * self._client.auto_refresh_delay), self.refresh_window)
-        return success
+            #if we need to set a refresh timer, do it:
+            is_hq = options.get("quality", 0)>=95
+            is_lossy = coding in ("jpeg", "vpx", "x264")
+            if self._refresh_timer is None and self._auto_refresh_delay>0 and is_lossy and not is_hq:
+                #make sure our own refresh does not make us fire again
+                #FIXME: this should be per-window!
+                if self._refresh_ignore_sequence<packet_sequence:
+                    #NOTE: for x264 and vpx, we always get full frames (whole window refresh)
+                    #this is not the case with jpeg but since jpeg does not switch the encoding on the fly, we're ok
+                    self._refresh_min_pixels = width*height
+                    self._refresh_ignore_sequence = packet_sequence+1
+                    self._refresh_timer = gobject.timeout_add(int(1000 * self._auto_refresh_delay), self.refresh_window)
+        callbacks.append(after_draw_refresh)
+        self._backing.draw_region(x, y, width, height, coding, img_data, rowstride, options, callbacks)
 
     """ gtk3 """
     def do_draw(self, context):
@@ -299,11 +435,12 @@ class ClientWindow(gtk.Window):
         self._backing.cairo_draw(context, x, y)
 
     def do_map_event(self, event):
-        log("Got map event")
+        log("Got map event: %s", event)
         gtk.Window.do_map_event(self, event)
         if not self._override_redirect:
             x, y, w, h = get_window_geometry(self)
-            self._client.send(["map-window", self._id, x, y, w, h])
+            client_properties = {"workspace" : self.get_workspace()}
+            self._client.send(["map-window", self._id, x, y, w, h, client_properties])
             self._pos = (x, y)
             self._size = (w, h)
         self._been_mapped = True
@@ -315,12 +452,15 @@ class ClientWindow(gtk.Window):
         if self._override_redirect:
             return
         x, y, w, h = get_window_geometry(self)
+        w = max(1, w)
+        h = max(1, h)
         ox, oy = self._pos
         dx, dy = x-ox, y-oy
         self._pos = (x, y)
         if self._client.window_configure:
             #if we support configure-window, send that first
-            self._client.send(["configure-window", self._id, x, y, w, h])
+            client_properties = {"workspace" : self.get_workspace()}
+            self._client.send(["configure-window", self._id, x, y, w, h, client_properties])
         if dx!=0 or dy!=0:
             #window has moved
             if not self._client.window_configure:
@@ -338,6 +478,8 @@ class ClientWindow(gtk.Window):
 
     def move_resize(self, x, y, w, h):
         assert self._override_redirect
+        w = max(1, w)
+        h = max(1, h)
         self.window.move_resize(x, y, w, h)
         self.new_backing(w, h)
 
