@@ -4,7 +4,7 @@
 # later version. See the file COPYING for details.
 
 ### NOTE: this must be kept in sync with the version in
-###    xpra/platform/gui.py 
+###    xpra/platform/gui.py
 #@PydevCodeAnalysisIgnore
 import os as _os
 import sys as _sys
@@ -15,9 +15,14 @@ elif _sys.platform.startswith("darwin"):
 elif _os.name == "posix":
     from xpra.xposix import *
 else:
-    raise OSError, "Unknown OS %s" % (_os.name)
+    raise OSError("Unknown OS %s" % (_os.name))
 
 def add_notray_option(parser, extra_text=""):
     parser.add_option("--no-tray", action="store_true",
                           dest="no_tray", default=False,
                           help="Disables the system tray%s" % extra_text)
+
+def add_delaytray_option(parser, extra_text=""):
+    parser.add_option("--delay-tray", action="store_true",
+                          dest="delay_tray", default=False,
+                          help="Waits for the first events before showing the system tray")
