@@ -4,7 +4,7 @@
 # later version. See the file COPYING for details.
 
 ### NOTE: this must be kept in sync with the version in
-###    xpra/platform/__init__.py 
+###    xpra/platform/__init__.py
 #@PydevCodeAnalysisIgnore
 import os as _os
 import sys as _sys
@@ -15,4 +15,4 @@ elif _sys.platform.startswith("darwin"):
 elif _os.name == "posix":
     from xpra.xposix.gui import *
 else:
-    raise OSError, "Unknown OS %s" % (_os.name)
+    raise OSError("Unknown OS %s" % (_os.name))
