@@ -1,5 +1,6 @@
 # This file is part of Parti.
 # Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
+# Copyright (C) 2011, 2012 Antoine Martin <antoine@nagafix.co.uk>
 # Parti is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -15,30 +16,33 @@ _display = gdk.get_display()
 assert _display, "cannot open the display with GTK, is DISPLAY set?"
 
 from xpra.platform.client_extras_base import ClientExtrasBase
+from xpra.platform.clipboard_base import DefaultClipboardProtocolHelper
 
 from wimpiggy.log import Logger
 log = Logger()
 
 
 class ClientExtras(ClientExtrasBase):
-    def __init__(self, client, opts):
-        ClientExtrasBase.__init__(self, client, opts)
+    def __init__(self, client, opts, conn):
+        ClientExtrasBase.__init__(self, client, opts, conn)
+        try:
+            from xpra.platform.gdk_clipboard import GDKClipboardProtocolHelper
+            self.setup_clipboard_helper(GDKClipboardProtocolHelper)
+        except ImportError, e:
+            log.error("GDK Clipboard failed to load: %s - using 'Default Clipboard' fallback", e)
+            self.setup_clipboard_helper(DefaultClipboardProtocolHelper)
         self.setup_menu(True)
-        self.setup_tray(opts.no_tray, opts.tray_icon)
+        self.setup_tray(opts.no_tray, opts.delay_tray, opts.tray_icon)
         self.setup_xprops(opts.pulseaudio)
         self.setup_x11_bell()
         self.has_dbusnotify = False
         self.has_pynotify = False
-        if not self.setup_dbusnotify() and not self.setup_pynotify():
-            log.error("turning notifications off")
-        try:
-            from xpra.xposix.xclipboard import ClipboardProtocolHelper
-            self.setup_clipboard_helper(ClipboardProtocolHelper)
-        except ImportError:
-            self.clipboard_helper = None
+        if opts.notifications:
+            if not self.setup_dbusnotify() and not self.setup_pynotify():
+                log.error("turning notifications off")
 
-    def exit(self):
-        ClientExtrasBase.exit(self)
+    def cleanup(self):
+        ClientExtrasBase.cleanup(self)
         if self.tray_widget:
             self.hide_tray()
             self.tray_widget = None
@@ -49,8 +53,10 @@ class ClientExtras(ClientExtrasBase):
         if sys.executable.startswith("/usr/local"):
             options.reverse()
         try:
-            # test for a local install (from do-build):
-            options.insert(0, os.path.join(os.path.dirname(sys.argv[0]), "..", "share", "xpra"))
+            # test for a local installation path (run from source tree):
+            local_share_path = os.path.join(os.path.dirname(sys.argv[0]), "..", "share", "xpra")
+            if os.path.exists(local_share_path):
+                options.insert(0, local_share_path)
         except:
             pass
         for x in options:
@@ -71,10 +77,23 @@ class ClientExtras(ClientExtrasBase):
             return  f
         return  None
 
-    def setup_statusicon(self, tray_icon_filename):
+    def setup_statusicon(self, delay_tray, tray_icon_filename):
         self.tray_widget = None
         try:
             self.tray_widget = gtk.StatusIcon()
+            def hide_tray(*args):
+                self.tray_widget.set_visible(False)
+            self.hide_tray = hide_tray
+            if delay_tray:
+                self.hide_tray()
+            def show_tray(*args):
+                log.info("showing tray")
+                #session_name will get set during handshake
+                self.tray_widget.set_visible(True)
+            if hasattr(self.tray_widget, "set_tooltip_text"):
+                self.tray_widget.set_tooltip_text(self.get_tray_tooltip())
+            else:
+                self.tray_widget.set_tooltip(self.get_tray_tooltip())
             self.tray_widget.connect('popup-menu', self.popup_menu)
             self.tray_widget.connect('activate', self.activate_menu)
             filename = self.get_tray_icon_filename(tray_icon_filename)
@@ -84,28 +103,25 @@ class ClientExtras(ClientExtrasBase):
                 else:
                     pixbuf = gdk.pixbuf_new_from_file(filename)
                     self.tray_widget.set_from_pixbuf(pixbuf)
-            def hide_tray(*args):
-                self.tray_widget.set_visible(False)
-            self.hide_tray = hide_tray
-            def show_tray(*args):
-                log.debug("showing tray")
-                #session_name will get set during handshake
-                self.tray_widget.set_visible(True)
-                if hasattr(self.tray_widget, "set_tooltip_text"):
-                    self.tray_widget.set_tooltip_text(self.client.session_name)
-                else:
-                    self.tray_widget.set_tooltip(self.client.session_name)
-            self.client.connect("handshake-complete", show_tray)
+            if delay_tray:
+                self.client.connect("first-ui-received", show_tray)
             return True
         except Exception, e:
-            log.debug("could not setup gtk.StatusIcon: %s", e)
+            log.info("could not setup gtk.StatusIcon: %s", e)
             return False
 
-    def setup_appindicator(self, tray_icon_filename):
+    def setup_appindicator(self, delay_tray, tray_icon_filename):
         try:
             import appindicator            #@UnresolvedImport
             filename = self.get_tray_icon_filename(tray_icon_filename)
             self.tray_widget = appindicator.Indicator("Xpra", filename, appindicator.CATEGORY_APPLICATION_STATUS)
+            def hide_appindicator(*args):
+                self.tray_widget.set_status(appindicator.STATUS_PASSIVE)
+            self.hide_tray = hide_appindicator
+            if delay_tray:
+                self.hide_tray()
+            def show_appindicator(*args):
+                self.tray_widget.set_status(appindicator.STATUS_ACTIVE)
             if hasattr(self.tray_widget, "set_icon_theme_path"):
                 self.tray_widget.set_icon_theme_path(self.get_icons_dir())
             self.tray_widget.set_attention_icon("xpra.png")
@@ -113,13 +129,9 @@ class ClientExtras(ClientExtrasBase):
                 self.tray_widget.set_icon(filename)
             else:
                 self.tray_widget.set_label("Xpra")
-            def hide_appindicator(*args):
-                self.tray_widget.set_status(appindicator.STATUS_PASSIVE)
-            self.hide_tray = hide_appindicator
-            def show_appindicator(*args):
-                self.tray_widget.set_status(appindicator.STATUS_ACTIVE)
             self.tray_widget.set_menu(self.menu)
-            self.client.connect("handshake-complete", show_appindicator)
+            if delay_tray:
+                self.client.connect("first-ui-received", show_appindicator)
             return  True
         except Exception, e:
             log.debug("could not setup appindicator: %s", e)
@@ -154,7 +166,7 @@ class ClientExtras(ClientExtrasBase):
         except:
             return False
 
-    def setup_tray(self, no_tray, tray_icon_filename):
+    def setup_tray(self, no_tray, delay_tray, tray_icon_filename):
         """ choose the most appropriate tray implementation
             Ubuntu is a disaster in this area, see:
             http://xpra.org/trac/ticket/43#comment:8
@@ -163,9 +175,9 @@ class ClientExtras(ClientExtrasBase):
         self.hide_tray = None
         if no_tray:
             return
-        if self._is_ubuntu_11_10_or_later() and self.setup_appindicator(tray_icon_filename):
+        if self._is_ubuntu_11_10_or_later() and self.setup_appindicator(delay_tray, tray_icon_filename):
             return
-        if not self.setup_statusicon(tray_icon_filename):
+        if not self.setup_statusicon(delay_tray, tray_icon_filename):
             log.error("failed to setup system-tray")
 
     def setup_dbusnotify(self):
@@ -204,7 +216,6 @@ class ClientExtras(ClientExtrasBase):
             log.error("cannot import x11 bell bindings (will use gtk fallback) : %s", e)
 
     def setup_xprops(self, pulseaudio):
-        self.client.connect("handshake-complete", self.setup_xprops)
         self.ROOT_PROPS = {
             "RESOURCE_MANAGER": "resource-manager"
             }
@@ -229,10 +240,12 @@ class ClientExtras(ClientExtrasBase):
 
     def _handle_xsettings_changed(self, *args):
         blob = self._xsettings_watcher.get_settings_blob()
+        log("xsettings_changed new value=%s", blob)
         if blob is not None:
             self.client.send(["server-settings", {"xsettings-blob": blob}])
 
     def _handle_root_prop_changed(self, obj, prop, value):
+        log("root_prop_changed: %s=%s", prop, value)
         assert prop in self.ROOT_PROPS
         if value is not None:
             self.client.send(["server-settings", {self.ROOT_PROPS[prop]: value.encode("utf-8")}])
@@ -326,8 +339,8 @@ class ClientExtras(ClientExtrasBase):
                 #ie: {"shift" : ["Shift_L", "Shift_R"], "mod1" : "Meta_L", ...]}
                 log.debug("modifier mappings=%s", mod_mappings)
                 meanings = {}
-                for modifier,keynames in mod_mappings.items():
-                    for keyname in keynames:
+                for modifier,keys in mod_mappings.items():
+                    for _,keyname in keys:
                         meanings[keyname] = modifier
                 return  meanings, [], []
         except ImportError, e:
@@ -356,6 +369,14 @@ class ClientExtras(ClientExtrasBase):
         log.debug("get_keymap_modifiers parsed: meanings=%s", meanings)
         return  meanings, [], []
 
+    def get_x11_keymap(self):
+        try:
+            from wimpiggy.lowlevel import get_keycode_mappings      #@UnresolvedImport
+            return get_keycode_mappings(gtk.gdk.get_default_root_window())
+        except Exception, e:
+            log.error("failed to use raw x11 keymap: %s", e)
+        return  ""
+
     def get_keymap_spec(self):
         xkbmap_print = self.exec_get_keyboard_data(["setxkbmap", "-print"])
         if xkbmap_print is None:
