@@ -1,3 +1,9 @@
+# This file is part of Parti.
+# Copyright (C) 2010-2012 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
+# Parti is released under the terms of the GNU GPL v2, or, at your option, any
+# later version. See the file COPYING for details.
+
 import gobject
 import gtk
 from wimpiggy.error import *
@@ -13,6 +19,7 @@ from wimpiggy.lowlevel import (
 from wimpiggy.log import Logger
 log = Logger()
 
+
 class XSettingsManager(object):
     def __init__(self, settings_blob):
         self._selection = ManagerSelection(gtk.gdk.display_get_default(),
@@ -66,7 +73,7 @@ class XSettingsWatcher(gobject.GObject):
     def do_wimpiggy_client_message_event(self, event):
         if (event.window is self._root
             and event.message_type == "MANAGER"
-            and event.data[1] == get_xatom(event.window, "_XSETTINGS_S0")):
+            and event.data[1] == get_xatom("_XSETTINGS_S0")):
             log("XSettings manager changed")
             self._add_watch()
             self.emit("xsettings-changed")
@@ -80,8 +87,7 @@ class XSettingsWatcher(gobject.GObject):
         owner = self._owner()
         if owner is None:
             return None
-        blob = prop_get(owner, "_XSETTINGS_SETTINGS", "xsettings-settings")
-        return blob
+        return prop_get(owner, "_XSETTINGS_SETTINGS", "xsettings-settings")
 
     def get_settings_blob(self):
         log("Fetching current XSettings data")
